/*-
 * %%----------------------------------------------------------------------------------------------
 * Solidify Framework - Solidify Frontend - user-preferences.util.ts
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2023 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

import {CookieType} from "../cookie-consent/enums/cookie-type.enum";
import {CookieConsentUtil} from "../cookie-consent/utils/cookie-consent.util";
import {LanguagePartialEnum} from "../enums/partial/language-partial.enum";
import {LocalStoragePartialEnum} from "../enums/partial/local-storage-partial.enum";
import {DefaultSolidifyEnvironment} from "../environments/environment.solidify-defaults";
import {LocalStorageHelper} from "../helpers/local-storage.helper";
import {
  isNotNullNorUndefined,
  isNullOrUndefined,
} from "../tools/is/is.tool";
import {ExtendEnum} from "../types/extend-enum.type";
import {SsrUtil} from "./ssr.util";

export class UserPreferencesUtil {
  static getPreferredLanguage(environment: DefaultSolidifyEnvironment): ExtendEnum<LanguagePartialEnum> {
    const savedLanguage = LocalStorageHelper.getItem(LocalStoragePartialEnum.language) as ExtendEnum<LanguagePartialEnum>;
    if (isNullOrUndefined(savedLanguage) || environment.appLanguages.indexOf(savedLanguage) === -1) {
      return this._determineDefaultLanguageAndStoreIntoLocalStorage(environment);
    }
    return savedLanguage;
  }

  private static _determineDefaultLanguageAndStoreIntoLocalStorage(environment: DefaultSolidifyEnvironment): ExtendEnum<LanguagePartialEnum> {
    const defaultLanguageDetermined = UserPreferencesUtil._determineDefaultLanguage(environment);
    if (CookieConsentUtil.isFeatureAuthorized(CookieType.localStorage, LocalStoragePartialEnum.language)) {
      LocalStorageHelper.setItem(LocalStoragePartialEnum.language, defaultLanguageDetermined);
    }
    return defaultLanguageDetermined;
  }

  private static _determineDefaultLanguage(environment: DefaultSolidifyEnvironment): ExtendEnum<LanguagePartialEnum> {
    const languageFromLocalBrowser = UserPreferencesUtil._findCorrespondingLanguageToLocal(environment);
    if (languageFromLocalBrowser === null) {
      return environment.defaultLanguage;
    }
    return languageFromLocalBrowser;
  }

  private static _findCorrespondingLanguageToLocal(environment: DefaultSolidifyEnvironment): ExtendEnum<LanguagePartialEnum> {
    const browserLanguage: string = SsrUtil.window?.navigator.language;

    const foundLanguage: ExtendEnum<LanguagePartialEnum> = environment.appLanguages.find(l => browserLanguage.startsWith(l));
    if (isNotNullNorUndefined(foundLanguage)) {
      return foundLanguage;
    }

    return null;
  }
}
